@@ -1292,7 +1292,7 @@ do_proc_packets( CTX c, IOBUF a )
             /* stop processing when an invalid packet has been encountered
              * but don't do so when we are doing a --list-packets. */
 	    if (gpg_err_code (rc) == GPG_ERR_INV_PACKET
-                && opt.list_packets != 2 )
+                && opt.list_packets == 0 )
 		break;
 	    continue;
 	}
